# -*- coding: utf-8 -*-

import json
import time
import platform
import socket
import requests
import psutil
import paho.mqtt.client as mqtt
import os
import threading
import uuid
import sys
import re
import subprocess
from datetime import datetime
if platform.system() == "Windows":
    import pythoncom
    import wmi


# Functie om huidige directory op te halen, ongeacht .py of .exe
def get_base_path():
    if getattr(sys, 'frozen', False):  # Wanneer het als .exe wordt uitgevoerd
        return os.path.dirname(sys.executable)
    else:
        return os.path.dirname(os.path.abspath(__file__))

# Configuratiepad instellen
config_path = os.path.join(get_base_path(), 'config.json')

# Configuratie laden uit config.json
with open(config_path, 'r') as f:
    config = json.load(f)
# Uniek systeem ID genereren
def generate_system_id():
    try:
        mac = hex(uuid.getnode()).replace('0x', '').upper()
        return mac
    except Exception as e:
        print("Fout bij genereren van MAC-gebaseerd ID: {}".format(e))
    return str(uuid.uuid1())

SYSTEM_ID = config.get("system_id") or generate_system_id()


# MQTT-server gegevens uit config
MQTT_BROKER = config["mqtt_broker"]
MQTT_PORT = config["mqtt_port"]
MQTT_USERNAME = config.get("mqtt_username")
MQTT_PASSWORD = config.get("mqtt_password")
MQTT_TOPIC = "{}/{}".format(config["mqtt_topic"], SYSTEM_ID)
MQTT_NOTIFY_TOPIC = "{}/{}".format(config["mqtt_notify_topic"], SYSTEM_ID)
MQTT_COMMAND_TOPIC = "{}/{}".format(config.get("mqtt_command_topic", "commands/topic"), SYSTEM_ID)
FRIENDLY_NAME = config["friendly_name"]
INTERVAL = config["interval_seconds"]






# Functie om meldingen te publiceren
def publish_message(topic, message):
    client.publish(topic, json.dumps(message))
    print("Melding gepubliceerd: {} - {}".format(topic, message))




def publish_service_start_notification():
    publish_message(MQTT_NOTIFY_TOPIC, {
        "event": "service_started",
        "timestamp": datetime.utcnow().isoformat()
    })
    print("Service gestart notificatie verzonden.")






def monitor_login_logout():
    previous_users = set(u.name for u in psutil.users())

    while True:
        if platform.system() == "Windows":
            # Windows-methode met psutil
            current_users = set(u.name for u in psutil.users())
        elif platform.system() == "Linux":
            # Linux-methode met `who`-commando om actieve sessies te detecteren
            output = subprocess.check_output("who", shell=True).decode("utf-8").splitlines()
            current_users = set(line.split()[0] for line in output if line.strip())

        # Detecteer nieuwe logins
        new_logins = current_users - previous_users
        if new_logins:
            for user in new_logins:
                publish_message(MQTT_NOTIFY_TOPIC, {
                    "event": "user_login",
                    "user": user,
                    "timestamp": datetime.utcnow().isoformat()
                })

        # Detecteer uitloggende gebruikers
        logged_out_users = previous_users - current_users
        if logged_out_users:
            for user in logged_out_users:
                publish_message(MQTT_NOTIFY_TOPIC, {
                    "event": "user_logout",
                    "user": user,
                    "timestamp": datetime.utcnow().isoformat()
                })

        # Update `previous_users` alleen als er veranderingen zijn
        if new_logins or logged_out_users:
            previous_users = current_users

        time.sleep(10)  # Interval om te controleren op wijzigingen

# Functie om lijst met geïnstalleerde programma's op te halen (Windows/Linux)
def get_installed_programs():
    system = platform.system()
    try:
        if system == "Windows":
            # Probeer de uitvoer te decoderen met 'latin-1' als 'utf-8' problemen geeft
            output = subprocess.check_output('wmic product get name', shell=True).decode('latin-1').splitlines()
            installed_programs = [line.strip() for line in output if line.strip() and 'Name' not in line]
        elif system == "Linux":
            try:
                try:
                    output = subprocess.check_output('dpkg -l', shell=True).decode('utf-8')
                except subprocess.CalledProcessError as e:
                    output = e.output.decode('utf-8')
                output = output.splitlines()
                installed_programs = [line.split()[1] for line in output if line.startswith("ii")]
            except subprocess.CalledProcessError:
                try:
                    output = subprocess.check_output('rpm -qa', shell=True).decode('utf-8')
                except subprocess.CalledProcessError as e:
                    output = e.output.decode('utf-8')
                output = output.splitlines()
                installed_programs = [line.strip() for line in output if line.strip()]

        else:
            print("Onbekend besturingssysteem. Kan geïnstalleerde programma's niet ophalen.")
            return set()
        
        return set(installed_programs)
    except Exception as e:
        print("Fout bij ophalen geïnstalleerde programma's:", e)
        return set()

        
# Monitor voor installatie/verwijdering van programma's
def monitor_installation_changes():
    previous_programs = get_installed_programs()
    
    while True:
        current_programs = get_installed_programs()
        
        # Nieuwe installaties detecteren
        new_programs = current_programs - previous_programs
        if new_programs:
            for program in new_programs:
                publish_message(MQTT_NOTIFY_TOPIC, {
                    "event": "new_application_installed",
                    "program": program,
                    "timestamp": datetime.utcnow().isoformat()
                })
                print("Nieuwe installatie gedetecteerd:", program)
        
        # Verwijderde programma's detecteren
        removed_programs = previous_programs - current_programs
        if removed_programs:
            for program in removed_programs:
                publish_message(MQTT_NOTIFY_TOPIC, {
                    "event": "application_removed",
                    "program": program,
                    "timestamp": datetime.utcnow().isoformat()
                })
                print("Verwijdering gedetecteerd:", program)
        
        previous_programs = current_programs
        time.sleep(60)

# Functie om schijfruimte te controleren en notificaties te versturen als de vrije ruimte <5% is
def monitor_disk_space():
    last_notification_date = {}  # Houdt de laatste meldingsdatum bij per partitie
    
    while True:
        # Controleer elke partitie
        for partition in psutil.disk_partitions():
            try:
                usage = psutil.disk_usage(partition.mountpoint)
                
                # Berekeningen in stappen uitvoeren om integerdeling te vermijden
                free_float = float(usage.free)
                total_float = float(usage.total)
                fraction_free = free_float / total_float if total_float > 0 else 0
                free_percent = round(fraction_free * 100, 2)

                # Controleer of vrije ruimte minder dan 5% is
                if free_percent < 5:
                    today = datetime.today().date()
                    
                    # Controleer of er al een melding voor deze partitie vandaag is verzonden
                    if last_notification_date.get(partition.device) != today:
                        publish_message(MQTT_NOTIFY_TOPIC, {
                            "event": "disk_space_low",
                            "message": "Schijf {} heeft nog maar {:.2f}% vrije ruimte over.".format(partition.device, free_percent)
                        })
                        # Update de laatste meldingsdatum voor deze partitie
                        last_notification_date[partition.device] = today
                        
            except Exception as e:
                print("Fout bij controle van schijf {}: {}".format(partition.device, e))

        # Controleer elke 60 seconden
        time.sleep(60)

# Controleer antivirus op Windows
def check_windows_antivirus():
    pythoncom.CoInitialize()  # Initialiseer COM-omgeving voor gebruik in threads
    try:
        # Verbinding met SecurityCenter2 namespace
        try:
            c = wmi.WMI(namespace="root\\SecurityCenter2")
            antiviruses = c.AntiVirusProduct()
        except Exception as e:
            print("Fout bij toegang tot SecurityCenter2:", e)
            c = wmi.WMI(namespace="root\\SecurityCenter")
            antiviruses = c.AntiVirusProduct()
        
        # Controleer antivirusstatus
        if antiviruses:
            status = "Virusscanner '{}' is geïnstalleerd en actief.".format(antiviruses[0].displayName)
        else:
            status = "Geen virusscanner gevonden."
    except Exception as e:
        status = "Fout bij antiviruscontrole op Windows: {}".format(e)
        print(status)
    finally:
        pythoncom.CoUninitialize()  # Sluit COM-omgeving af
    return status

# Controleer antivirus op Linux (ClamAV)
def check_linux_antivirus():
    try:
        try:
            output = subprocess.check_output("systemctl is-active clamd", shell=True).decode("utf-8")
        except subprocess.CalledProcessError as e:
            output = e.output.decode("utf-8")
        if output.strip() == "active":
            return "ClamAV is actief."
        else:
            return "ClamAV is niet actief."
    except subprocess.CalledProcessError:
        return "ClamAV is niet geïnstalleerd of niet actief."


# Bepaal het besturingssysteem en controleer de antivirusstatus
def get_antivirus_status():
    os_type = platform.system()
    if os_type == "Windows":
        return check_windows_antivirus()
    elif os_type == "Linux":
        return check_linux_antivirus()
    else:
        return "Onbekend besturingssysteem. Antiviruscontrole niet uitgevoerd."

# Functie om de antivirusstatus regelmatig te controleren en alleen te melden bij veranderingen
def monitor_antivirus_status(interval=600):
    # Haal de initiële status op, zonder een melding te sturen
    last_status = get_antivirus_status()

    while True:
        # Controleer de huidige antivirusstatus
        current_status = get_antivirus_status()

        # Verstuur alleen een melding als de status is veranderd
        if current_status != last_status:
            publish_message(MQTT_NOTIFY_TOPIC, {
                "event": "antivirus_status",
                "message": current_status,
                "timestamp": datetime.utcnow().isoformat()
            })
            last_status = current_status  # Werk de laatste status bij

        # Wacht het opgegeven interval
        time.sleep(interval)


# Controleer Windows Firewall
def check_windows_firewall():
    try:
        output = subprocess.check_output("netsh advfirewall show allprofiles", shell=True).decode()
        
        # Controleer de status van elk profiel door de uitvoer op te splitsen
        domain_on = "Domain Profile Settings" in output and "State                                 ON" in output.split("Domain Profile Settings")[1]
        private_on = "Private Profile Settings" in output and "State                                 ON" in output.split("Private Profile Settings")[1]
        public_on = "Public Profile Settings" in output and "State                                 ON" in output.split("Public Profile Settings")[1]

        # Retourneer de firewallstatus
        if domain_on and private_on and public_on:
            return "Windows Firewall is ingeschakeld voor alle profielen."
        elif domain_on or private_on or public_on:
            return "Windows Firewall is ingeschakeld voor ten minste één profiel."
        else:
            return "Windows Firewall is uitgeschakeld voor alle profielen."
    except Exception as e:
        print("Fout bij controle van Windows Firewall:", e)
        return None

# Controleer UFW Firewall (Linux)
def check_linux_firewall():
    try:
        try:
            output = subprocess.check_output("ufw status", shell=True).decode("utf-8")
        except subprocess.CalledProcessError as e:
            output = e.output.decode("utf-8")
        return "Status: active" in output
    except subprocess.CalledProcessError:
        return None

# Controleer Firewalld (Linux)
def check_firewalld():
    try:
        try:
            output = subprocess.check_output("systemctl is-active firewalld", shell=True).decode("utf-8")
        except subprocess.CalledProcessError as e:
            output = e.output.decode("utf-8")
        return output.strip() == "active"
    except subprocess.CalledProcessError:
        return None

# Functie om firewallstatus regelmatig te controleren en wijzigingen te melden
def monitor_firewall_status():
    # Haal de initiële firewallstatus op zonder een melding te sturen
    if platform.system() == "Windows":
        last_status = check_windows_firewall()
    elif platform.system() == "Linux":
        try:
            subprocess.check_output("ufw status", shell=True)
            last_status = check_linux_firewall()
        except subprocess.CalledProcessError:
            last_status = check_firewalld()
    else:
        print("Onbekend besturingssysteem; firewallstatus niet gecontroleerd.")
        return  # Stop de functie als het OS onbekend is

    # Start de monitoring loop
    while True:
        # Controleer de huidige firewallstatus opnieuw
        if platform.system() == "Windows":
            status = check_windows_firewall()
        elif platform.system() == "Linux":
            try:
                subprocess.check_output("ufw status", shell=True)
                status = check_linux_firewall()
            except subprocess.CalledProcessError:
                status = check_firewalld()
        else:
            print("Onbekend besturingssysteem; firewallstatus niet gecontroleerd.")
            break

        # Verstuur alleen een melding als de firewallstatus is veranderd na de eerste iteratie
        if status is not None and status != last_status:
            publish_message(MQTT_NOTIFY_TOPIC, {
                "event": "firewall_status",
                "message": status,
                "timestamp": datetime.utcnow().isoformat()
            })
            last_status = status  # Werk de laatste status bij

        # Wacht 5 minuten (300 seconden) voordat de volgende controle wordt uitgevoerd
        time.sleep(300)
        




# Controle voor shutdown/reboot processen voor zowel Windows als Linux
def monitor_shutdown_command():
    while True:
        # Detecteer shutdown/reboot voor Windows
        if platform.system() == "Windows":
            process_names = ['shutdown.exe', 'reboot']
        
        # Detecteer shutdown/reboot voor Linux
        elif platform.system() == "Linux":
            process_names = ['shutdown', 'reboot', 'poweroff']
            
            # Controleer systemd geplande shutdown op sommige Linux-systemen
            if os.path.exists('/run/systemd/shutdown/scheduled'):
                publish_message("system_shutdown_scheduled")
                time.sleep(10)  # Wacht om duplicaten te voorkomen

        # Controleer actieve processen
        for proc in psutil.process_iter(['name']):
            try:
                if proc.info['name'] in process_names:
                    publish_message("system_shutdown_or_reboot_started")
                    time.sleep(10)  # Wacht om duplicaten te voorkomen
                    break
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue

        time.sleep(5)

def publish_system_start_notification():
    publish_message(MQTT_NOTIFY_TOPIC, {
        "event": "system_started",
        "timestamp": datetime.utcnow().isoformat()
    })
    print("Systeem opgestart notificatie verzonden.")

# Mapping van Windows-versies naar gebruiksvriendelijke namen
WINDOWS_VERSION_NAMES = {
    "10": "Windows 10",
    "11": "Windows 11",
    "8.1": "Windows 8.1",
    "8": "Windows 8",
    "7": "Windows 7",
    "Vista": "Windows Vista",
    "XP": "Windows XP",
    "2016Server": "Windows Server 2016",
    "2019Server": "Windows Server 2019",
    "2022Server": "Windows Server 2022",
    "2008ServerR2": "Windows Server 2008 R2",
    "2008Server": "Windows Server 2008",
    "2003Server": "Windows Server 2003"
}

def get_os_name_and_version():
    os_name = platform.system()
    os_version = platform.release() if os_name != "Windows" else WINDOWS_VERSION_NAMES.get(platform.release(), "Onbekende versie")
    return os_name, os_version

def on_connect(client, userdata, flags, rc):
    print("Verbonden met resultaatcode {}".format(rc))
    client.subscribe(MQTT_COMMAND_TOPIC)

def on_message(client, userdata, msg):
    print("Bericht ontvangen op topic {}: {}".format(msg.topic, msg.payload.decode()))
    try:
        command = json.loads(msg.payload.decode())
    except json.JSONDecodeError:
        print("Ongeldig JSON-formaat")
        return
    
    if "action" not in command:
        print("Geen actie gevonden in bericht")
        return
    
    action = command["action"].lower()
    if action == "restart":
        restart_system()
    elif action == "shutdown":
        shutdown_system()
    elif action == "logout":
        username = command.get("username")
        if username:
            logout_user(username)
        else:
            print("Gebruikersnaam niet opgegeven voor afmelden")
    elif action == "message":
        console_message = command.get("message")
        if console_message:
            print("Bericht: {}".format(console_message))
        else:
            print("Geen bericht opgegeven voor weergave")

def restart_system():
    print("Systeem wordt opnieuw opgestart...")
    os.system("shutdown /r /t 1" if platform.system() == "Windows" else "sudo reboot")

def shutdown_system():
    print("Systeem wordt uitgeschakeld...")
    os.system("shutdown /s /t 1" if platform.system() == "Windows" else "sudo shutdown now")


def logout_user(username):
    print("Afmelden van gebruiker: {}".format(username))
    
    if platform.system() == "Windows":
        try:
            # Initialiseer WMI-client voor Windows
            c = wmi.WMI()

            # Zoek naar actieve sessies van de opgegeven gebruiker
            sessions = c.Win32_LogonSession()
            for session in sessions:
                for user in session.references("Win32_LoggedOnUser"):
                    account = user.Antecedent
                    if account.Name.lower() == username.lower():
                        # Sessie gevonden, gebruik shutdown.exe om de gebruiker af te melden
                        subprocess.check_call("shutdown.exe /l", shell=True)
                        print("Gebruiker {} is afgemeld.".format(username))
                        return
            print("Geen actieve sessie gevonden voor gebruiker {}.".format(username))
        
        except Exception as e:
            print("Fout bij het afmelden van gebruiker {}: {}".format(username, e))
    
    elif platform.system() == "Linux":
        try:
            # Gebruik pkill om de gebruiker af te melden op Linux
            os.system("pkill -KILL -u {}".format(username))
            print("Gebruiker {} is afgemeld.".format(username))
        except Exception as e:
            print("Fout bij het afmelden van gebruiker: {}".format(e))
    else:
        print("Afmelden wordt niet ondersteund op dit besturingssysteem.")

def get_system_info():
    try:
        # Systeeminformatie
        uptime = int(time.time() - psutil.boot_time())
        cpu_usage = psutil.cpu_percent(interval=1)
        cpu_info = platform.processor()
        cpu_cores = psutil.cpu_count(logical=False)
        os_name, os_version = get_os_name_and_version()
        hostname = socket.gethostname()
        internal_ip = socket.gethostbyname(hostname)
        total_memory = round(psutil.virtual_memory().total / (1024 * 1024))  # In MB
        free_memory = round(psutil.virtual_memory().available / (1024 * 1024))  # In MB
        external_ip = requests.get('https://api.ipify.org').text
        users = [{"name": u.name, "terminal": u.terminal, "started": u.started} for u in psutil.users()]
        timestamp = datetime.utcnow().isoformat() + "Z"  # ISO 8601-formaat in UTC

        # Firewall- en antivirusstatus als True/False
        if platform.system() == "Windows":
            firewall_status = check_windows_firewall()
            antivirus_status = check_windows_antivirus()
        elif platform.system() == "Linux":
            try:
                subprocess.check_output("ufw status", shell=True)
                firewall_status = check_linux_firewall()
            except subprocess.CalledProcessError:
                firewall_status = check_firewalld()
            antivirus_status = check_linux_antivirus()
        else:
            firewall_status = False
            antivirus_status = False

        # Vrije schijfruimte per partitie
        disk_info = []
        for partition in psutil.disk_partitions():
            try:
                usage = psutil.disk_usage(partition.mountpoint)
                
                # Berekeningen in stappen uitvoeren voor compatibiliteit met Python 2.7
                total_space_mb = usage.total / (1024 * 1024)
                free_space_mb = usage.free / (1024 * 1024)

                # Zet zowel `usage.free` als `usage.total` om naar float om integerdeling te vermijden
                free_float = float(usage.free)
                total_float = float(usage.total)
                fraction_free = free_float / total_float if total_float > 0 else 0
                percent_free = fraction_free * 100

                disk_info.append({
                    "device": partition.device,
                    "mountpoint": partition.mountpoint,
                    "total_space_mb": total_space_mb,
                    "free_space_mb": free_space_mb,
                    "percent_free": round(percent_free, 2)  # Vrije ruimte in procenten
                })


                
            except PermissionError:
                # Als we geen toegang hebben tot een bepaalde partitie, slaan we deze over
                continue

        # Processen informatie
        processes = []
        for proc in psutil.process_iter(attrs=['pid', 'name', 'status']):
            try:
                processes.append(proc.info)
                if len(processes) >= 5:
                    break
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue

        # Services informatie
        services = []
        if platform.system() == "Windows":
            for service in psutil.win_service_iter():
                try:
                    svc_name = service.name()
                    svc_status = service.status()
                    svc_display_name = service.display_name()
                    services.append({
                        "name": svc_name,
                        "display_name": svc_display_name,
                        "status": svc_status,
                    })
                    if len(services) >= 20: 
                        break
                except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.Error, Exception) as e:
                    print("Fout bij ophalen serviceinformatie voor service {}: {}".format(service.name(), e))
                    continue
        else:
            if os.system("which systemctl > /dev/null") == 0:
                output = os.popen("systemctl list-units --type=service --no-pager --all").read()
                for line in output.splitlines():
                    if "loaded" in line:
                        parts = line.split()
                        services.append({
                            "name": parts[0],
                            "status": parts[3] if len(parts) > 3 else "unknown",
                        })
                        if len(services) >= 20:
                           break

        # Gedetailleerde netwerkinterface-informatie
        network_info = {}
        for interface, addrs in psutil.net_if_addrs().items():
            iface_info = {"IPv4": None, "IPv6": None, "MAC": None}
            for addr in addrs:
                if addr.family == socket.AF_INET:
                    iface_info["IPv4"] = addr.address
                    iface_info["netmask"] = addr.netmask
                    iface_info["broadcast"] = addr.broadcast
                elif addr.family == socket.AF_INET6:
                    iface_info["IPv6"] = addr.address
                    iface_info["netmask_v6"] = addr.netmask
                elif addr.family == psutil.AF_LINK:
                    iface_info["MAC"] = addr.address
            
            stats = psutil.net_if_stats().get(interface)
            if stats:
                iface_info["is_up"] = stats.isup
                iface_info["speed_mbps"] = stats.speed
                iface_info["duplex"] = stats.duplex
                iface_info["mtu"] = stats.mtu

            network_info[interface] = iface_info

        # Return alle verzamelde informatie inclusief firewall, antivirus, en schijfruimte-informatie
        return {
            "system_id": SYSTEM_ID,
            "uptime": uptime,
            "friendly_name": FRIENDLY_NAME,
            "cpu_usage": cpu_usage,
            "cpu_info": cpu_info,
            "cpu_cores": cpu_cores,
            "os_name": os_name,
            "os_version": os_version,
            "internal_ip": internal_ip,
            "total_memory_mb": total_memory,
            "free_memory_mb": free_memory,
            "external_ip": external_ip,
            "hostname": hostname,
            "logged_in_users": users,
            "processes": processes,
            "services": services,
            "network_info": network_info,
            "firewall_status": bool(firewall_status),
            "antivirus_status": bool(antivirus_status),
            "disk_info": disk_info,  # Informatie over schijfruimte
            "timestamp": timestamp
        }
    except Exception as e:
        print("Fout bij ophalen systeeminfo: {}".format(e))
        return None


def publish_system_info():
    while True:
        system_info = get_system_info()
        if system_info:
            message = json.dumps(system_info)
            client.publish(MQTT_TOPIC, message)
            print("Bericht gepubliceerd: {}".format(message))
        time.sleep(INTERVAL)

client = mqtt.Client()
if MQTT_USERNAME and MQTT_PASSWORD:
    client.username_pw_set(MQTT_USERNAME, MQTT_PASSWORD)
client.on_connect = on_connect
client.on_message = on_message
client.connect(MQTT_BROKER, MQTT_PORT, 60)

# Start een thread voor het verzenden van systeeminformatie
info_thread = threading.Thread(target=publish_system_info)
info_thread.daemon = True
info_thread.start()

login_logout_thread = threading.Thread(target=monitor_login_logout)
login_logout_thread.daemon = True
login_logout_thread.start()


# Start een aparte thread voor reboot/shutdown monitoring
reboot_shutdown_thread = threading.Thread(target=monitor_shutdown_command)
reboot_shutdown_thread.daemon = True
reboot_shutdown_thread.start()

# Start de installatie/verwijdering monitoring in een aparte thread
installation_monitor_thread = threading.Thread(target=monitor_installation_changes)
installation_monitor_thread.daemon = True
installation_monitor_thread.start()

# Start de schijfruimte monitoring in een aparte thread
disk_monitor_thread = threading.Thread(target=monitor_disk_space)
disk_monitor_thread.daemon = True
disk_monitor_thread.start()


# Start de firewall monitoring in een aparte thread
firewall_monitor_thread = threading.Thread(target=monitor_firewall_status)
firewall_monitor_thread.daemon = True
firewall_monitor_thread.start()

# Start de antivirusmonitoring in een aparte thread
antivirus_monitor_thread = threading.Thread(target=monitor_antivirus_status)
antivirus_monitor_thread.daemon = True
antivirus_monitor_thread.start()

# Stuur een notificatie zodra de service start
publish_service_start_notification()
# stuur een notificatie als de server opgestart is
publish_system_start_notification()

# MQTT-client in een continue loop om berichten te verwerken
client.loop_forever()




